#! /usr/bin/env python
# ______________________________________________________________________________
"""jbmesher

This script is intended to produce a Dolfin mesh for a journal bearing.

usage:
  -h    Print this usage message
  -X <val> x coordinate of the journal center
  -Y <val> y coordinate of the journal center
  -R <val> radius of journal
  -H <val> refinement of boundary. (point every sin(x+H),cos(x+H))
"""

# _____________________________________________________________________________
# Module Copyright and license
__author__ = "Andy R Terrel (aterrel@uchicago.edu)"
__date__ = "2008-09"
__copyright__ = "Copyright (C) 2008 Andy R Terrel"
__license__ = "GNU LGPL Version 3.0 or later"

# _____________________________________________________________________________
# Module Imports
import sys
import os
import getopt
from math import sqrt, pi, sin, cos

# ______________________________________________________________________________
# Module Functions
def createGEO(R,X,Y,LC,FILE):
    EPS = 1e-8
    small_dx = 1.0 - R - sqrt(X**2 + Y**2)
    big_dx = 1.0 - R + sqrt(X**2 + Y**2)
    LC_small = LC
    LC_big = LC * big_dx/small_dx

    points1 = []
    points2 = []
    t = 0.0
    points1.append( (0, 0, LC) )
    points2.append( (X, Y, LC) )
    while t < 2*pi - EPS:
        out_x = sin(t)
        out_y = cos(t)
        in_x = R*sin(t)+X
        in_y = R*cos(t)+Y
        dx = 1.0 - sqrt(in_x**2 + in_y**2)
        p_LC = LC * dx/big_dx
        points1.append( (sin(t), cos(t), p_LC) )
        points2.append( (R*sin(t)+X, R*cos(t)+Y, p_LC*R) )
        t += pi/2
    points = points1+points2

    geo_fp = open(FILE+".geo","w")

    for i,pt in enumerate(points):
        geo_fp.write("Point(%d) = {%f,%f,0.0,%f};\n" % (i+1,pt[0], pt[1], pt[2]))

    # PT 1,2,3,4  5,6,7,8
    # Lines 1 (1,2); 2(2,3); 3(3,4); 4(4,1)
    #       5 (5,6); 6(6,7); 7(7,8); 8(8,5)
    #
    #
    #
    j = 0
    for i in xrange(2,5):
        j += 1
        geo_fp.write("Circle(%d) = {%d, %d, %d};\n" % (j, i, 1, i+1))
    j += 1
    geo_fp.write("Circle(%d) = {%d, %d, %d};\n" % (j, 5, 1, 2))

    for i in xrange(7,10):
        j += 1
        geo_fp.write("Circle(%d) = {%d, %d, %d};\n" % (j, i, 6, i+1))
    j += 1
    geo_fp.write("Circle(%d) = {%d, %d, %d};\n" % (j, 10, 6, 7))

    j += 1
    geo_fp.write("Line Loop(%d) = {%s};\n" %
                 (j,",".join(map(str, range(1,5)))))
    j += 1
    geo_fp.write("Line Loop(%d) = {%s};\n" %
                 (j, ",".join(map(str,range(5,9)))))
    j += 1
    geo_fp.write("Plane Surface(%d) = {%d,%d};\n" % (j, j-2, j-1))
    geo_fp.close()

# _____________________________________________________________________________
def main(*args):
    R = 0.9
    X = 0.0
    Y = 0.0
    H = None
    try:
        opts, remainder = getopt.getopt(args, "R:X:Y:H:h")
    except getopt.GetoptError:
        print "Unrecognized options"
        print __doc__
        sys.exit(1)
    for opt_key, opt_val in opts:
        if opt_key == "-R":
            R = float(opt_val)
        elif opt_key == "-X":
            X = float(opt_val)
        elif opt_key == "-Y":
            Y = float(opt_val)
        elif opt_key == "-H":
            H = float(opt_val)
        elif opt_key == "-h":
            print __doc__
    if not H:
        LC = (1-max(R+abs(X),R+abs(Y)))/2
    else:
        LC = H

    FILE = "jb_r_%.2f_x_%.2f_y_%.2f_h_%.3f" %(R,X,Y,H)
    FILE = FILE.replace(".","_")
    createGEO(R,X,Y,H,FILE)
    os.system("gmsh %s.geo -2" % FILE)
    os.system("dolfin-convert %s.msh %s.xml" % (FILE,FILE))
    return

# ______________________________________________________________________________
# Module Script Runner
if __name__ == "__main__":
    main(*sys.argv[1:])
# ______________________________________________________________________________
# End of mesher.py
