// Copyright (C) 2008-2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// Modified by Patrick Riesen, 2009
//
// First added:  2008-09-10
// Last changed: 2009-12-03
//
// This demo implements a journal bearing


#include <rheagen.h>

using namespace dolfin;


// ----------------------------------------------------------------------------
// Boundary Condition for outer wall
class NoSlipBC : public Expression
{
public:
  NoSlipBC() : Expression(2) {}
  void eval (double* values, const std::vector<double>& x) const
  {
      values[0] = 0;
      values[1] = 0;
  }
};

// ----------------------------------------------------------------------------
// Boundary Condition for journal
class JournalBC : public Expression
{
private:
  const double radius;
  const double x_val;
  const double y_val;
public:
  JournalBC(const double radius, const double x_val, const double y_val):
    Expression(2), radius(radius), x_val(x_val), y_val(y_val)
  {}
  void eval (double* values, const std::vector<double>& x) const
  {
    values[0] = (x[1] - y_val)/radius;
    values[1] = -(x[0] - x_val)/radius;
  }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain
class CenterCircle : public SubDomain
{
private:
  const double radius;
  const double x_val;
  const double y_val;
public:
  CenterCircle(const double radius, const double x_val, const double y_val) :
    SubDomain(), radius(radius), x_val(x_val), y_val(y_val)
  {
    // Do nothing
  }
  virtual bool inside (const double* x, bool on_boundary) const
  {
    double dbco = (1-sqrt(x_val*x_val+y_val*y_val)-radius);
    double dfjc = sqrt((x[0]-x_val)*(x[0]-x_val) + (x[1]-y_val)*(x[1]-y_val));
    bool ret_val = on_boundary && dfjc < radius + dbco/2;
    return on_boundary && ret_val;
  }
};

class OuterCircle : public SubDomain
{
private:
  const double radius;
  const double x_val;
  const double y_val;
public:
  OuterCircle(const double radius, const double x_val, const double y_val) :
    SubDomain(), radius(radius), x_val(x_val), y_val(y_val)
  {
    // Do nothing
  }
  virtual bool inside (const double* x, bool on_boundary) const
  {
    double dbco = (1-sqrt(x_val*x_val+y_val*y_val)-radius);
    double dfjc = sqrt((x[0]-x_val)*(x[0]-x_val) + (x[1]-y_val)*(x[1]-y_val));
    bool ret_val = dfjc > radius + dbco/2;
    return on_boundary && ret_val;
  }
};

// ----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
  char meshfile[200] = "../data/meshes/journalbearing/jb_r_0_90_x_0_00_y_-0_05_h_0_050.xml.gz";
  double center_radius = 0.90;
  double center_x = 0.0;
  double center_y = -0.05;


  // Create the mesh, define the subsystems, and mark the boundaries.
  // by the position the bc function is in the bc func array.
  Mesh mesh(meshfile);
  mesh.order();
  CenterCircle center(center_radius,center_x,center_y);
  OuterCircle outer(center_radius,center_x,center_y);
  MeshFunction<unsigned int> vel_subdomains(mesh, mesh.topology().dim()-1);
  vel_subdomains = 2;
  center.mark(vel_subdomains,0);
  outer.mark(vel_subdomains,1);
  JournalBC   journal_bc(center_radius, center_x, center_y);
  NoSlipBC    no_slip_bc;
  std::vector< GenericFunction* > vel_bc_funcs;
  vel_bc_funcs.push_back(&journal_bc);
  vel_bc_funcs.push_back(&no_slip_bc);

  // RHS Function
  Constant zero(2, 0.0);

  // Create fluid and solver
  Fluid fluid(mesh, vel_subdomains, vel_bc_funcs);

  Grade2Solver grade2;
  grade2.parameters["max_iter"] = 15;
  grade2.parameters["alpha"] = 0.1;
  grade2.parameters["mu"] = 0.5;
  grade2.solve(fluid, zero);
  fluid.toFile("grade2");

  // Perhaps use a stokes solver instead
  StokesSolver stokes;
  stokes.solve(fluid, zero);
  fluid.toFile("stokes");


  // Perhaps use a power law stokes solver instead
  PowerlawStokesSolver powstokes;
  powstokes.parameters["powerlaw_exponent"] = 2.5;
  powstokes.parameters["continuation_steps"] = 1;
  powstokes.solve(fluid, zero);
  fluid.toFile("powerlaw_stokes");

  return 0;
}

// ----------------------------------------------------------------------------
// end of file
