// Copyright (C) 2008-2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// Modified by Patrick Riesen, 2009
//
// First added:  2003-11-28
// Last changed: 2009-12-08
//
// This demo implements a simple lid driven cavity with a parabolic driving
// function on the lid.

#include <rheagen.h>

using namespace dolfin;

// ----------------------------------------------------------------------------
// Boundary Subdomain marking walls
class Wall : public SubDomain
{
public:
    bool inside (const double *x, bool on_boundary) const
  {
    return on_boundary;
  }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain marking lid
class Lid : public SubDomain
{
public:
    bool inside (const double *x, bool on_boundary) const
  {
    return on_boundary && (x[1] > 0.75 - DOLFIN_EPS);
  }
};

// ----------------------------------------------------------------------------
// Function for top flow for velocity
class TopFlow : public Expression
{
public:
  TopFlow() :
    Expression(2)
  {}
  void eval (double* values, const std::vector<double>& x) const
  {
    values[0] = 4.0 * x[0] * (1.0 - x[0]);
    values[1] = 0.0;
  }
};

// Function for no-slip boundary condition for velocity
class Noslip : public Expression
{
public:
  Noslip() :
    Expression(2)
  {}
  void eval(double* values, const std::vector<double>& x) const
  {
    values[0] = 0.0;
    values[1] = 0.0;
  }
};
// -----------------------------------------------------------------------------
int main ()
{

  // Create a unit square
  UnitSquare mesh(32, 32, "crossed");
  Wall wall;
  Lid lid;

  // No-slip boundary condition for velocity
  Noslip noslip;

  // Lid boundary condition for velocity
  TopFlow topflow;

  // Create a mesh function, marking boundaries by the
  // position the bc function is in the bc func array.
  MeshFunction<unsigned int> vel_subdomains(mesh, mesh.topology().dim() - 1);
  vel_subdomains = 2;
  wall.mark(vel_subdomains, 0);
  lid.mark(vel_subdomains, 1);
  std::vector< GenericFunction* > vel_bc_funcs;
  vel_bc_funcs.push_back(&noslip);
  vel_bc_funcs.push_back(&topflow);

  // Create fluid and solvers
  Fluid fluid(mesh, vel_subdomains, vel_bc_funcs);
  Constant zero(2, 0.0);

  Grade2Solver grade2;
  grade2.parameters["max_iter"] = 15;
  grade2.parameters["alpha"] = 0.1;
  grade2.parameters["mu"] = 10.0;
  grade2.solve(fluid, zero);
  fluid.toFile("grade2");

  // Perhaps use a stokes solver instead
  StokesSolver stokes;
  stokes.parameters["velocity_disc"] = "Taylor-Hood";
  stokes.solve(fluid, zero);
  fluid.toFile("stokes");


  // Perhaps use another stokes solver instead
  StokesSolver stokes2;
  stokes2.parameters["velocity_disc"] = "Scott-Vogelius";
  stokes2.parameters["SV_max_iter"] = 5;
  stokes2.solve(fluid, zero);
  fluid.toFile("stokes_SV");


  // Or Powerlaw Solver
  PowerlawStokesSolver powstokes;
  powstokes.parameters["powerlaw_exponent"] = 1.5;
  powstokes.solve(fluid, zero);
  fluid.toFile("powerlaw_stokes");

  return 0;
}

// ----------------------------------------------------------------------------
// end of file
