// Copyright (C) 2008-2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:  2008-11-19
// Last changed: 2009-12-08
//
// This demo implements a 4 to 1 contraction with symmetric domain.

#include <rheagen.h>

using namespace dolfin;


// ----------------------------------------------------------------------------
// Boundary Condition for top wall
class NoSlipBC : public Expression
{
public:
  NoSlipBC() : Expression(2) {}
  void eval (double* values, const std::vector<double>& input) const
  {
      values[0] = 0;
      values[1] = 0;
  }
};
// ----------------------------------------------------------------------------
// Boundary Condition for inflow
class Inflow : public Expression
{
public:
  Inflow() : Expression(2) {}
  void eval (double* values, const std::vector<double>& input) const
  {
    double y = input[1];
    values[0] = 1.0/64.0 * (4.0 - y) * (4.0 + y);
    values[1] = 0.0;
  }
};

class Outflow : public Expression
{
 public:
  Outflow() : Expression(2) {}
  void eval (double* values, const std::vector<double>& input) const
  {
    double y = input[1];
    values[0] = (1.0 - y) * (1.0 + y);
    values[1] = 0.0;
  }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain
class Inlet : public SubDomain
{
 public:
  Inlet () : SubDomain() {}
  virtual bool inside (const double* pt, bool on_boundary) const
  {
    return on_boundary && pt[0] < DOLFIN_SQRT_EPS;
  }
};

class Outlet : public SubDomain
{
 public:
  Outlet () : SubDomain() {}
  virtual bool inside (const double* pt, bool on_boundary) const
  {
    return on_boundary && pt[0] > 6.0 - DOLFIN_SQRT_EPS;
  }
};

class TopWall : public SubDomain
{
public:
  TopWall() : SubDomain() {}
  virtual bool inside (const double* pt, bool on_boundary) const
  {
    double x = pt[0], y = pt[1];
    return on_boundary &&
      ( ( x > 3.0 - DOLFIN_SQRT_EPS && y > 1.0 - DOLFIN_SQRT_EPS ) ||
        ( x < 3.0 + DOLFIN_SQRT_EPS && y > 4.0 - DOLFIN_SQRT_EPS));
  }
};

class SymmetryLine: public SubDomain
{
public:
  SymmetryLine() : SubDomain() {}
  virtual bool inside (const double* pt, bool on_boundary) const
  {
    return on_boundary && pt[1] < DOLFIN_SQRT_EPS;
  }
};

// ----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
  char meshfile[200] = "../data/meshes/planar_contraction/pc_h_0_050.xml.gz";

  // Create the mesh, define the subsystems, and mark the boundaries.
  // by the position the bc function is in the bc func array.
  Mesh mesh(meshfile);
  mesh.order();
  mesh.init();
  Inlet inlet;
  Outlet outlet;
  TopWall top_wall;
  SymmetryLine sym_line;
  MeshFunction<unsigned int> vel_subdomains(mesh, mesh.topology().dim()-1);
  vel_subdomains = 4;
  top_wall.mark(vel_subdomains, 0);
  sym_line.mark(vel_subdomains, 1);
  inlet.mark(vel_subdomains, 2);
  outlet.mark(vel_subdomains, 3);
  Inflow inflow;
  Outflow outflow;
  NoSlipBC no_slip_bc;
  Constant sym_bc(0.0);
  std::vector< GenericFunction* > vel_bc_funcs;
  vel_bc_funcs.push_back(&no_slip_bc);
  vel_bc_funcs.push_back(&sym_bc);
  vel_bc_funcs.push_back(&inflow);
  vel_bc_funcs.push_back(&outflow);
  // Specify which component to apply velocity function,
  //   -1 all, [0, 1, 2] x, y, z components
  std::vector< int > vel_components;
  vel_components.push_back(-1);
  vel_components.push_back(1);
  vel_components.push_back(-1);
  vel_components.push_back(-1);

  // RHS Function
  Constant zero(2, 0.0);

  // Create fluid and solver
  Fluid fluid(mesh, vel_subdomains, vel_bc_funcs, vel_components);

  Grade2Solver grade2;
  grade2.parameters["max_iter"] = 15;
  grade2.parameters["alpha"] = 10.0;
  grade2.parameters["mu"] = 0.5;
  grade2.solve(fluid, zero);
  fluid.toFile("grade2");

  // Perhaps use a stokes solver instead
  StokesSolver stokes;
  stokes.solve(fluid, zero);
  fluid.toFile("stokes");

  return 0;
}

// ----------------------------------------------------------------------------
// end of file
