#include <dolfin.h>
#include "OB_DEVSS_SU.h"
#include "OB_StrainProj.h"
#include "OB_Stab.h"

using namespace dolfin;


// Classes for marking pressure conditions (if necessary)
class PinPoint : public SubDomain
{
public:
  bool inside (const double *x, bool on_boundary) const
  { return fabs(x[1]  < DOLFIN_EPS) && fabs(x[0] < DOLFIN_EPS); }
};

class PressurePin : public Expression
{
public:
  PressurePin() : Expression() {}
  void eval (double* values, const std::vector<double>& x) const
  { values[0] = 0.0; }
};

class Lam : public Expression
{
public:
  Lam(double lam) : Expression(), _lam(lam) {}
  void eval(double* vals, const std::vector<double>& input) const
  {
    vals[0] =_lam;
  }
  inline void setLam(double new_lam) {_lam = new_lam;}
private:
  double _lam;
};

// ----------------------------------------------------------------------------
// Boundary Subdomain marking walls
class Wall : public SubDomain
{
public:
    bool inside (const double *x, bool on_boundary) const
  { return on_boundary; }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain marking lid
class Lid : public SubDomain
{
public:
    bool inside (const double *x, bool on_boundary) const
  { return on_boundary && (x[1] > 1.0 - DOLFIN_EPS); }
};

// ----------------------------------------------------------------------------
// Function for top flow for velocity
class TopFlow : public Expression
{
public:
  TopFlow() : Expression(2) {}
  void eval (double* values, const std::vector<double>& x) const
  { values[0] = 4.0 * x[0] * (1.0 - x[0]); values[1] = 0.0; }
};

// Function for no-slip boundary condition for velocity
class Noslip : public Expression
{
public:
  Noslip() : Expression(2) {}
  void eval(double* values, const std::vector<double>& x) const
  { values[0] = 0.0; values[1] = 0.0; }
};

class OB_DEVSS_SU_Solver : public NonlinearProblem
{
public:
  OB_DEVSS_SU_Solver(Parameters p) :
    NonlinearProblem(), dt(p["dt"]), T(p["T"]), lam_fin(p["lam"]),
    eta(p["eta"]), eta_e(p["eta_e"]), n(p["n"])
  {
    // Create a unit square
    mesh = UnitSquare(n, n, "crossed");
    f = new Constant(2, 0.0);

    // Create function spaces
    _fs = new OB_DEVSS_SU::FunctionSpace(mesh);
    _fs_vel = new SubSpace(*_fs, 0);
    _fs_press = new SubSpace(*_fs, 1);
    _fs_proj = new OB_StrainProj::FunctionSpace(mesh);
    _fs_stab = new OB_Stab::FunctionSpace(mesh);

    // Create BCS
    bcs.push_back( new DirichletBC(*_fs_vel, noslip, wall) );
    bcs.push_back( new DirichletBC(*_fs_vel, topflow, lid) );
    bcs.push_back( new DirichletBC(*_fs_press, pressurepin, pinpoint, "pointwise") );

    // Create parameter functions
    f_dt = new Constant(dt);
    f_lam = new Lam(0.0);
    f_eta = new Constant(eta);
    f_eta_e = new Constant(eta_e);
    f_h = new CellSize(mesh);

    // Create discrete functions
    _uu_vec = new Vector(_fs->dofmap().global_dimension());
    _uu = new Function(*_fs, *_uu_vec);
    _D_h_vec = new Vector(_fs_proj->dofmap().global_dimension());
    _D_h = new Function(*_fs_proj, *_D_h_vec); //(*_fs_proj);
    _uu0_vec = _uu_vec;// new Vector(_fs_proj->dofmap().global_dimension());
    _uu0 = new Function(*_fs, *_uu0_vec);
    _stab_vec = new Vector(_fs_stab->dofmap().global_dimension());
    _stab = new Function(*_fs_stab, *_stab_vec);
    //    _stab->vector(); // make _stab create its own vector

    // Set forms
    _a = new OB_DEVSS_SU::BilinearForm(*_fs, *_fs);
    _a->sol = *_uu;
    //    _a->sol0 = *_uu0;
    _a->dt = *f_dt;
    _a->lam = *f_lam;
    _a->eta = *f_eta;
    _a->eta_e = *f_eta_e;
    _a->su = *_stab;

    _L = new OB_DEVSS_SU::LinearForm(*_fs);
    _L->sol = *_uu;
    _L->sol0 = *_uu0;
    _L->dt = *f_dt;
    _L->lam = *f_lam;
    _L->eta = *f_eta;
    _L->eta_e = *f_eta_e;
    _L->f = *f;
    _L->D_proj_vec = *_D_h;
    _L->su = *_stab;

    _u_vec = new Vector(_fs_vel->dofmap().global_dimension());
    _u = new Function(*_fs_vel, *_u_vec);
    _a_proj = new OB_StrainProj::BilinearForm(*_fs_proj, *_fs_proj);
    _L_proj = new OB_StrainProj::LinearForm(*_fs_proj);
    _L_proj->u = *_u;

    _a_stab = new OB_Stab::BilinearForm(*_fs_stab, *_fs_stab);
    _L_stab = new OB_Stab::LinearForm(*_fs_stab);
    _L_stab->u = *_u;
    _L_stab->h = *f_h;

    // Create assembler and matrices
    _A_proj = new Matrix;
    _b_proj = new Vector;
    _A_stab = new Matrix;
    _b_stab = new Vector;

    _newton_solver.parameters["maximum_iterations"] = 20;

  }

  void F(GenericVector &b, const GenericVector& x)
  {
    *_u = (*_uu)[0];

    logging(false);
    assemble(b, *_L);

    for (unsigned int i = 0; i < bcs.size(); i++)
      bcs[i]->apply(b, x);
    logging(true);
  }

  void J(GenericMatrix& A, const GenericVector& x)
  {
    logging(false);
    assemble(A, *_a);
    for (unsigned int i = 0; i < bcs.size(); i++)
      bcs[i]->apply(A);
    assemble(*_b_proj, *_L_proj);
    _lu_solver.solve(*_A_proj, _D_h->vector(), *_b_proj);
    assemble(*_b_stab, *_L_stab);
    _lu_solver.solve(*_A_stab, _stab->vector(), *_b_stab);
    logging(true);
  }

  void solve()
  {

  // Store variables for time and lambda stepping
  double lam_curr = 0.0;
  double d_lam = 0.001;
  double lam_sgn = 1.0 ? lam_fin >= 0.0 : -1.0;

  // Assemble strain projection matrix and stabilization term
  logging(false);
  assemble(*_A_proj, *_a_proj);
  assemble(*_A_stab, *_a_stab);
  logging(true);


  begin("Starting Oldroyd-B Solver.");
  while (lam_curr < lam_fin + DOLFIN_EPS) {
    info("-----------------------");
    info("Lambda: %.2e <= %.2e", lam_curr, lam_fin);
    f_lam->setLam(lam_curr);

    begin("Starting Time stepper.");
    for (double t=0.0; t < T; t += dt) {
      info("--------------------------");
      info("T: %.3f", t);

      // Call the newton solver
      _newton_solver.solve(*this, _uu->vector());

      // Check to see if the time step has converged to a steady state
      if (_newton_solver.iteration() < 2) break;
      _uu0_vec = _uu_vec;
    }
    end();

    // Double lambda or 0.1
    d_lam = lam_sgn * std::max(lam_curr, 0.1);
    if (lam_curr > lam_fin - DOLFIN_EPS)
      lam_curr = lam_fin + DOLFIN_EPS;
    else
      lam_curr = lam_sgn * std::min(fabs(lam_fin), fabs(lam_curr + d_lam));
  }
  end();
  }

private:
  // Constants
  double dt, T, lam_fin, eta, eta_e;
  int n;


  //Domain info
  Mesh mesh;
  Wall wall;
  Lid lid;
  Noslip noslip;
  TopFlow topflow;
  PinPoint pinpoint;
  PressurePin pressurepin;
  std::vector<const BoundaryCondition* > bcs;
  Constant* f;
  CellSize* f_h;

  //Function spaces
  FunctionSpace* _fs;
  SubSpace* _fs_vel;
  SubSpace* _fs_press;
  FunctionSpace* _fs_proj;
  FunctionSpace* _fs_stab;
  Vector* _uu_vec;
  Function* _uu;
  Vector* _u_vec;
  Function* _u;
  Vector* _uu0_vec;
  Function* _uu0;
  Vector* _D_h_vec;
  Function* _D_h;
  GenericFunction* f_dt;
  GenericFunction* f_eta;
  GenericFunction* f_eta_e;
  Lam* f_lam;
  Vector* _stab_vec;
  Function* _stab;


  OB_DEVSS_SU::BilinearForm* _a;
  OB_DEVSS_SU::LinearForm* _L;
  OB_StrainProj::BilinearForm* _a_proj;
  OB_StrainProj::LinearForm* _L_proj;
  OB_Stab::BilinearForm* _a_stab;
  OB_Stab::LinearForm* _L_stab;


  Matrix* _A_proj;
  Vector* _b_proj;
  Matrix* _A_stab;
  Vector* _b_stab;

  NewtonSolver _newton_solver;
  LUSolver _lu_solver;

};


int main(int argc, char* argv[])
{
  Parameters p("app_params");
  p.add("n", 16);
  p.add("dt", 1e-4);
  p.add("T", 100e-4);
  p.add("lam", 1.0);
  p.add("eta", 1.0);
  p.add("eta_e", 1.0);
  p.parse(argc, argv);

  info(p, true);
  cout << endl;

  OB_DEVSS_SU_Solver solver(p);
  solver.solve();
}
