#include <dolfin.h>
#include "Stokes.h"

using namespace dolfin;


// Classes for marking pressure conditions (if necessary)
class PinPoint : public SubDomain
{
public:
  bool inside (const double *x, bool on_boundary) const
  { return fabs(x[1]  < DOLFIN_EPS) && fabs(x[0] < DOLFIN_EPS); }
};

class PressurePin : public Expression
{
public:
  PressurePin() : Expression() {}
  void eval (double* values, const std::vector<double>& x) const
  { values[0] = 0.0; }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain marking walls
class Wall : public SubDomain
{
public:
    bool inside (const double *x, bool on_boundary) const
  { return on_boundary; }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain marking lid
class Lid : public SubDomain
{
public:
    bool inside (const double *x, bool on_boundary) const
  { return on_boundary && (x[1] > 1.0 - DOLFIN_EPS); }
};

// ----------------------------------------------------------------------------
// Function for top flow for velocity
class TopFlow : public Expression
{
public:
  TopFlow() : Expression(2) {}
  void eval (double* values, const std::vector<double>& x) const
  { values[0] = 4.0 * x[0] * (1.0 - x[0]); values[1] = 0.0; }
};

// Function for no-slip boundary condition for velocity
class Noslip : public Expression
{
public:
  Noslip() : Expression(2) {}
  void eval(double* values, const std::vector<double>& x) const
  { values[0] = 0.0; values[1] = 0.0; }
};

class StokesSolver : public NonlinearProblem
{
public:
  StokesSolver()
  {
    // Create a unit square
    mesh = UnitSquare(16, 16, "crossed");
    f = new Constant(2, 0.0);

    // Create function spaces
    _fs = new Stokes::FunctionSpace(mesh);
    _fs_vel = new SubSpace(*_fs, 0);
    _fs_press = new SubSpace(*_fs, 1);

    // Create BCS
    bcs.push_back( new DirichletBC(*_fs_vel, noslip, wall) );
    bcs.push_back( new DirichletBC(*_fs_vel, topflow, lid) );
    bcs.push_back( new DirichletBC(*_fs_press, pressurepin, pinpoint, "pointwise") );

    // Create discrete functions
    _uu_vec = new Vector(_fs->dofmap().global_dimension());
    _uu = new Function(*_fs, *_uu_vec);

    // Set forms
    _a = new Stokes::BilinearForm(*_fs, *_fs);
    _L = new Stokes::LinearForm(*_fs);
    std::map<std::string, const GenericFunction*> coeffs;
    coeffs["f"] = f;
    coeffs["sol"] = _uu;
    _L->set_coefficients(coeffs);

    // Create assembler and matrices
    _assembler = new Assembler;

  }

  void F(GenericVector &b, const GenericVector& x)
  {
    _assembler->assemble(b, *_L);

    for (unsigned int i = 0; i < bcs.size(); i++)
      bcs[i]->apply(b, x);
  }

  void J(GenericMatrix& A, const GenericVector& x)
  {
    _assembler->assemble(A, *_a);
    for (unsigned int i = 0; i < bcs.size(); i++)
      bcs[i]->apply(A);
  }

  void solve()
  {
    _newton_solver.solve(*this, *_uu_vec);
  }

private:
  // Constants
  double dt, T, lam_fin;

  //Domain info
  Mesh mesh;
  Wall wall;
  Lid lid;
  Noslip noslip;
  TopFlow topflow;
  PinPoint pinpoint;
  PressurePin pressurepin;
  std::vector<const BoundaryCondition* > bcs;
  Constant* f;

  //Function spaces
  FunctionSpace* _fs;
  SubSpace* _fs_vel;
  SubSpace* _fs_press;

  //Functions
  Vector* _uu_vec;
  Function* _uu;

  Form* _a;
  Form* _L;

  Assembler* _assembler;
  NewtonSolver _newton_solver;

};


int main()
{
  StokesSolver ss;
  ss.solve();
}
