// Copyright (C) 2008--2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:   2008-09-10
// Last modified: 2009-12-08

#include "Fluid.h"

using namespace dolfin;

//-----------------------------------------------------------------------------
Fluid::Fluid (Mesh &mesh) :
  _mesh(mesh)
{
  zeroPtrs();
}
//-----------------------------------------------------------------------------
Fluid::Fluid (Mesh& mesh, MeshFunction<uint>& vel_subdomains,
	      std::vector< GenericFunction* >& vel_bc_funcs) :
  _mesh(mesh)
{
  zeroPtrs();
  _vel_subdomains = &vel_subdomains;
  std::vector< int > vel_components;
  for (uint i = 0; i < vel_bc_funcs.size(); i++)
    vel_components.push_back(-1);

  regBCs(vel_bc_funcs, vel_components);
}
//-----------------------------------------------------------------------------
Fluid::Fluid (Mesh& mesh, MeshFunction<uint>& vel_subdomains,
	      std::vector< GenericFunction* >& vel_bc_funcs,
              std::vector< int >& vel_components) :
  _mesh(mesh)
{
  zeroPtrs();
  _vel_subdomains = &vel_subdomains;
  regBCs(vel_bc_funcs, vel_components);
}
// ----------------------------------------------------------------------------
Fluid::Fluid (const Fluid& that) :
  _mesh(that._mesh)
{
  throw std::runtime_error("Fluid copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
Fluid::~Fluid ()
{
  deletePtrs();
}
// ----------------------------------------------------------------------------
Fluid& Fluid::operator= (const Fluid& that)
{
  throw std::runtime_error("Fluid assignment is not implemented.");
  return *this;
}
//-----------------------------------------------------------------------------
void Fluid::setVelFunspace (FunctionSpace& vel_funspace)
{
  delete _vel_funspace;
  _vel_funspace = new FunctionSpace(vel_funspace);
}
//-----------------------------------------------------------------------------
void Fluid::setPressFunspace (FunctionSpace& press_funspace)
{
  delete _press_funspace;
  _press_funspace = new FunctionSpace(press_funspace);
}
//-----------------------------------------------------------------------------
void Fluid::setStressFunspace (FunctionSpace& stress_funspace)
{
  delete _stress_funspace;
  _stress_funspace = new FunctionSpace(stress_funspace);
}
//-----------------------------------------------------------------------------
void Fluid::init ()
{
  delete _velocity; _velocity = 0;
  if (_vel_funspace == 0)
    throw std::runtime_error("Fluid must have velocity function space");
  else
    _velocity = new Function(*_vel_funspace);

  delete _pressure; _pressure = 0;
  if (_press_funspace != 0)
    _pressure = new Function(*_press_funspace);

  delete _stress; _stress = 0;
  if (_press_funspace != 0)
    _stress = new Function(*_press_funspace);
}
//-----------------------------------------------------------------------------
void Fluid::regBCs (std::vector< GenericFunction* >& vel_bc_funcs,
                    std::vector< int >& vel_components)
{
  if (vel_bc_funcs.size() != vel_components.size())
    throw std::runtime_error("Unequal number of velocity components and functions");
  // FIXME: Need to create subspaces when we know the function spaces
  for (uint i = 0; i < vel_bc_funcs.size(); i++) {
    _vel_bc_funcs.push_back(vel_bc_funcs[i]);
    _vel_bc_components.push_back(vel_components[i]);
  }
}
//-----------------------------------------------------------------------------
void Fluid::toFile (std::string base_filename, bool print_pressure)
{
  std::string velocity_filename = base_filename + "_velocity.pvd";
  File velocity_writer(velocity_filename.c_str());
  velocity_writer << *_velocity;
  if (print_pressure) {
    std::string pressure_filename = base_filename + "_pressure.pvd";
    File pressure_writer(pressure_filename.c_str());
    pressure_writer << *_pressure;
  }
}
// ----------------------------------------------------------------------------
std::vector<const BoundaryCondition* > Fluid::createBCs (FunctionSpace& funspace,
                                                         bool mixed_space,
                                                         bool pressure_pin)
{
  std::vector<const BoundaryCondition* > bcs;

  for (int i=0; i < 2; i++)
    _vel_subspaces.push_back(new SubSpace(*_vel_funspace, i) );

  for (uint i = 0; i < _vel_bc_funcs.size(); i++){
    if (_vel_bc_components[i] == -1) {
      bcs.push_back( new DirichletBC(*_vel_funspace, *_vel_bc_funcs[i],
                                     *_vel_subdomains, i) );
    } else {
      bcs.push_back( new DirichletBC(*_vel_subspaces[_vel_bc_components[i]],
                                     *_vel_bc_funcs[i],
                                     *_vel_subdomains, i) );
    }
  }

  if (pressure_pin) {
    _pinpoint = new PinPoint();
    _pressurepin = new PressurePin();
    // Add a pressure pinpoint to the boundary conditions
    bcs.push_back( new DirichletBC(*_press_funspace, *_pressurepin,
                                   *_pinpoint, "pointwise") );
  }

  return bcs;
}
// ----------------------------------------------------------------------------
void Fluid::destroyBCs (std::vector<const BoundaryCondition* >& bcs)
{
  for (uint i = 0; i < bcs.size(); i++)
    delete bcs[i];
  bcs.resize(0);
}
// ----------------------------------------------------------------------------
void Fluid::deletePtrs ()
{
  delete _velocity;
  delete _pressure;
  delete _stress;
  for (uint i = 0; i < _vel_subspaces.size(); i++) {
    delete _vel_subspaces[i];
  }
  _vel_subspaces.resize(0);
  delete _vel_funspace;
  delete _press_funspace;
  delete _stress_funspace;
  delete _pinpoint;
  delete _pressurepin;
  zeroPtrs();
}
//-----------------------------------------------------------------------------
void Fluid::zeroPtrs ()
{
  _velocity = 0;
  _pressure = 0;
  _stress = 0;
  _vel_funspace = 0;
  _press_funspace = 0;
  _stress_funspace = 0;
  _pinpoint = 0;
  _pressurepin = 0;
}
//-----------------------------------------------------------------------------
// end of file
