// Copyright (C) 2008--2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:   2008-09-10
// Last modified: 2009-12-08

#ifndef __FLUID_H
#define __FLUID_H

#include <dolfin.h>

namespace dolfin
{

  typedef unsigned int uint;

  // Class representing the fluid
  class Fluid
  {
  public:
    Fluid (Mesh& mesh);

    Fluid (Mesh& mesh, MeshFunction<uint>& vel_subdomains,
	   std::vector< GenericFunction* >& vel_bc_funcs);

    Fluid (Mesh& mesh, MeshFunction<uint>& vel_subdomains,
	   std::vector< GenericFunction* >& vel_bc_funcs,
           std::vector<int>& vel_components);

    Fluid (const Fluid& that);

    ~Fluid ();

    Fluid& operator= (const Fluid& that);

    void init ();

    void regBCs (std::vector< GenericFunction* >& vel_bc_funcs,
                 std::vector<int>& vel_components);


    void toFile (std::string base_filename, bool print_pressure=true);

    std::vector<const BoundaryCondition* > createBCs (FunctionSpace& funspace,
                                                      bool mixed_space=true,
                                                      bool pressure_pin=false);

    void destroyBCs (std::vector<const BoundaryCondition* >& bcs);

    inline Mesh& mesh () {return _mesh;}
    inline Function& velocity () {return *_velocity;}
    inline Function& pressure () {return *_pressure;}
    inline Function& stress () {return *_stress;}
    inline FunctionSpace* vel_funspace () {return _vel_funspace;}
    inline FunctionSpace* press_funspace () {return _press_funspace;}
    inline FunctionSpace* stress_funspace () {return _stress_funspace;}
    inline MeshFunction<uint>* vel_subdomains () {return _vel_subdomains;}
    inline std::vector< GenericFunction* > vel_bc_funcs () {return _vel_bc_funcs;}
    inline std::vector< int > vel_bc_components () {return _vel_bc_components;}

    void setVelFunspace (FunctionSpace& vel_funspace);
    void setPressFunspace (FunctionSpace& press_funspace);
    void setStressFunspace (FunctionSpace& stress_funspace);

  protected:

    // Classes for marking pressure conditions (if necessary)
    class PinPoint : public SubDomain
    {
    public:
      PinPoint () : SubDomain(), pinpoint_set(false)
      { pinpoint[0] = 0.0; pinpoint[1] = 0.0; }

      bool inside (const double *x, bool on_boundary) const
      {
        if (pinpoint_set) {
          return (fabs(x[1] - pinpoint[1]) < DOLFIN_EPS) &&
            (fabs(x[0] - pinpoint[0]) < DOLFIN_EPS);
        } else {
          pinpoint_set = true;
          pinpoint[0] = x[0];
          pinpoint[1] = x[1];
          return true;
        }
      }
    private:
      mutable bool pinpoint_set;
      mutable double pinpoint[2];
    };

    class PressurePin : public Expression
    {
    public:
      PressurePin() : Expression() {}
      void eval (double* values, const std::vector<double>& x) const
      { values[0] = 0.0; }
    };

    void zeroPtrs ();
    void deletePtrs ();

    Mesh& _mesh;

    Function* _velocity;
    Function* _pressure;
    Function* _stress;
    MeshFunction<uint>* _vel_subdomains;
    std::vector< GenericFunction* > _vel_bc_funcs;
    std::vector< int > _vel_bc_components;
    FunctionSpace* _vel_funspace;
    std::vector< SubSpace* > _vel_subspaces;
    FunctionSpace* _press_funspace;
    FunctionSpace* _stress_funspace;
    PinPoint* _pinpoint;
    PressurePin* _pressurepin;
  };

}
#endif
