// Copyright (C) 2008-2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:  2008-09-12
// Last changed: 2009-12-08

#include "Grade2Solver.h"

using namespace dolfin;

// ----------------------------------------------------------------------------
Grade2Solver::Grade2Solver () :
  StokesSolver()
{
  zeroPtrs();
  // Set default parameters
  parameters.add("alpha", 1.0);
  parameters.add("mu", 1.0);
  parameters.add("max_iter",10);
}
// ----------------------------------------------------------------------------
Grade2Solver::Grade2Solver (const Grade2Solver& that)
{
  throw std::runtime_error("Copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
Grade2Solver::~Grade2Solver ()
{
  StokesSolver::deletePtrs();
  deletePtrs();
}
// ----------------------------------------------------------------------------
Grade2Solver& Grade2Solver::operator= (const Grade2Solver& that)
{
  throw std::runtime_error("Assignment is not implemented.");
  return *this;
}
// ----------------------------------------------------------------------------
void Grade2Solver::formFactory (Fluid& fluid, GenericFunction& f)
{
  // Remove old forms and functions
  deletePtrs();

  // Create necessary functions
  f_alpha = new Constant(parameters["alpha"]);
  f_mu = new Constant(parameters["mu"]);
  f_h = new CellSize(fluid.mesh());

  // Create the forms and initialize necessary functions
  const std::string vel_disc = parameters["velocity_disc"];
  if (vel_disc == "Taylor-Hood") {
    _funspace = new P2THG2Vel::FunctionSpace(fluid.mesh());
    _trans_funspace = new P2THG2Trans::FunctionSpace(fluid.mesh());
    _stab_funspace = new P2THG2Stab::FunctionSpace(fluid.mesh());
    SubSpace vel_fs(*_funspace, 0);
    SubSpace press_fs(*_funspace, 1);
    fluid.setVelFunspace(vel_fs);
    fluid.setPressFunspace(press_fs);
    fluid.init();
    _uu = new Function(*_funspace);
    _z = new Function(*_trans_funspace);
    _tau = new Function(*_stab_funspace);

    _a_stokes = new P2THG2Vel::BilinearForm(*_funspace, *_funspace,
                                            *_z, *f_mu);
    _L_stokes = new P2THG2Vel::LinearForm(*_funspace, f);

    _a_transport = new P2THG2Trans::BilinearForm(*_trans_funspace,
                                                 *_trans_funspace,
                                                 fluid.velocity(),
                                                 *f_mu, *f_alpha, *_tau);
    _L_transport = new P2THG2Trans::LinearForm(*_trans_funspace,
                                               fluid.velocity(),
                                              *f_mu, *f_alpha, *_tau, f);

    _a_stab = new P2THG2Stab::BilinearForm(*_stab_funspace, *_stab_funspace);
    _L_stab = new P2THG2Stab::LinearForm(*_stab_funspace,
                                         fluid.velocity(), *f_h);

    _L_res = new P2THG2Res::LinearForm(*_funspace,
                                       fluid.velocity(), fluid.pressure(),
                                       *_z, *f_mu, f);

  } else {
    throw  std::runtime_error("Unimplemented Grade2 velocity discretization");
  }
}
// ----------------------------------------------------------------------------
void Grade2Solver::solve (Fluid& fluid, GenericFunction& f, bool verbose)
{
  const std::string vel_disc = parameters["velocity_disc"];

  // Files for writing stuff
  File vel_file("results/velocity.pvd");
  File vel_diff_file("results/velocity_diff.pvd");
  File press_file("results/pressure.pvd");
  File press_diff_file("results/press_diff.pvd");
  File z_file("results/transport.pvd");
  File z_diff_file("results/transport_diff.pvd");

  // Generate the forms
  formFactory(fluid, f);

  // Create boundary conditions on the velocity and pressure
  std::vector<const BoundaryCondition* > bcs;
  if (vel_disc == "Taylor-Hood")
    bcs = fluid.createBCs(*_funspace, true, true);
  else if (vel_disc == "Scott-Vogelius")
    bcs = fluid.createBCs(*_funspace, false);

  // Prepare solver
  Assembler assembler;
  LUSolver solver;
  Matrix A, Az, Atau;
  Vector b, bz, btau;
  GenericVector& x = _uu->vector();
  GenericVector& xz = _z->vector();
  GenericVector& xtau = _tau->vector();

  double* tmp = new double[xtau.size()];
  uint* tmp_rows = new uint[xtau.size()];
  for (uint i = 0; i < xtau.size(); i++) tmp_rows[i] = i;
  double alpha = parameters["alpha"];

  // Iterate between Stokes and Transport solvers until converged
  double converged = 1.0;
  int max_iter = parameters["max_iter"];
  double r0_norm = 0.0;
  double x0_norm = 0.0;
  double z0_norm = 0.0;
  Vector x0, z0;
  Vector dx, dz, r_stokes;
  double rel_dx, rel_dz, rel_res;

  info("Starting grade-2 solver");
  for (int iter = 0; iter < max_iter && converged > DOLFIN_SQRT_EPS;
       iter++) {
    if (vel_disc == "Taylor-Hood") {
      logging(false);
      // Solve for (u,p)
      if (iter == 0)
        assembler.assemble(A, *_a_stokes);
      else
        assembler.assemble(A, *_a_stokes, false);
      assembler.assemble(b, *_L_stokes);
      for (uint i = 0; i < bcs.size(); i++)
        bcs[i]->apply(A, b);
      solver.solve(A, x, b);

      // Extract subfunctions
      fluid.velocity() = (*_uu)[0];
      fluid.pressure() = (*_uu)[1];

      // Compute stabilization term
      if (fabs(alpha) > DOLFIN_EPS) {
        if (iter == 0)
          assembler.assemble(Atau, *_a_stab);
        else
          assembler.assemble(Atau, *_a_stab, false);
        assembler.assemble(btau, *_L_stab);
        solver.solve(Atau, xtau, btau);
        xtau.get(tmp, xtau.size(), tmp_rows);
        for(uint i = 0; i < xtau.size(); ++i) {
          if(tmp[i] > DOLFIN_SQRT_EPS)
            tmp[i] = 2.0/(fabs(alpha)*sqrt(tmp[i]));
          else
            tmp[i] = 1.0/fabs(alpha);
        }
        xtau.set(tmp, xtau.size(), tmp_rows);
      }

      if (iter == 0)
        assembler.assemble(Az, *_a_transport);
      else
        assembler.assemble(Az, *_a_transport, false);
      assembler.assemble(bz, *_L_transport);
      solver.solve(Az, xz, bz);

      // Check for convergence
      assembler.assemble(r_stokes, *_L_res);
      Vector bctmp(r_stokes.size());
      bctmp.zero();
      for (dolfin::uint i = 0; i < bcs.size(); i++) {
        bcs[i]->apply(A, bctmp);
        bcs[i]->apply(r_stokes, bctmp);
      }
      if (iter == 0) {
        r0_norm = r_stokes.norm("l2");
        x0_norm = x.norm("l2");
        x0.resize(x.size());
        x0.zero();
        z0_norm = xz.norm("l2");
        z0.resize(xz.size());
        z0.zero();
      }

      // Compute dx = x_{n+1} - x_{n}
      dx = x; dx -= x0; x0 = x;
      dz = xz; dz -= z0; z0 = xz;
      rel_res = r_stokes.norm("l2")/r0_norm;
      //rel_res = 0.0;
      rel_dx = dx.norm("l2")/x0_norm;
      rel_dz = dz.norm("l2")/z0_norm;
      converged = rel_dx + rel_dz;

      logging(true);
      info("Iteration = %d, ||r||/||r_0||, dx/dx0, dz/dz0 = %.2e, %.2e, %.2e",
              iter, rel_res, rel_dx, rel_dz);

    /*
    } else if (vel_disc == "Scott-Vogelius") {
      scottVogeliusSolve(fluid,bcs,f,*f_w);
    */
    } else {
      throw std::runtime_error("Undefined velocity solver");
    }

    if (verbose) {
      // Write functions to file
      vel_file << fluid.velocity();
      press_file << fluid.pressure();
      z_file << *_z;

      // Write projection of error to file
      if (vel_disc == "Taylor-Hood") {
        P2VecDiff::FunctionSpace fs_v_diff(fluid.mesh());
	P2VecDiff::BilinearForm a_v_diff(fs_v_diff, fs_v_diff);
	P2VecDiff::LinearForm L_v_diff(fs_v_diff, fluid.velocity(), *_v_ex);
	VariationalProblem pde_v_diff(a_v_diff, L_v_diff);
	Function fun_v_diff(fs_v_diff);
	pde_v_diff.solve(fun_v_diff);
	vel_diff_file << fun_v_diff;

        P1Diff::FunctionSpace fs_p_diff(fluid.mesh());
	P1Diff::BilinearForm a_p_diff(fs_p_diff, fs_p_diff);
	P1Diff::LinearForm L_p_diff(fs_p_diff, fluid.pressure(), *_p_ex);
	VariationalProblem pde_p_diff(a_p_diff, L_p_diff);
	Function fun_p_diff(fs_p_diff);
	pde_p_diff.solve(fun_p_diff);
	press_diff_file << fun_p_diff;

        P2Diff::FunctionSpace fs_z_diff(fluid.mesh());
	P2Diff::BilinearForm a_z_diff(fs_z_diff, fs_z_diff);
	P2Diff::LinearForm L_z_diff(fs_z_diff, *_z, *_z_ex);
	VariationalProblem pde_z_diff(a_z_diff, L_z_diff);
	Function fun_z_diff(fs_z_diff);
	pde_z_diff.solve(fun_z_diff);
	z_diff_file << fun_z_diff;
      }
    }
  }

  // Clean up
  fluid.destroyBCs(bcs);
  delete tmp;
  delete tmp_rows;
}
// ----------------------------------------------------------------------------
void Grade2Solver::solve (Fluid& fluid, GenericFunction& f,
			  GenericFunction& v_ex, GenericFunction& p_ex,
                          GenericFunction& z_ex)
{
  _v_ex = &v_ex;
  _p_ex = &p_ex;
  _z_ex = &z_ex;
  solve(fluid, f, true);
}
// ----------------------------------------------------------------------------
void Grade2Solver::deletePtrs ()
{
  delete _uu;
  delete _z;
  delete f_alpha;
  delete f_mu;
  delete f_rho;
  delete f_w;
  delete f_h;
  delete _trans_funspace;
  delete _stab_funspace;
  delete _a_transport;
  delete _L_transport;
  delete _a_stab;
  delete _L_stab;
  delete _L_res;
  zeroPtrs();
}
// ----------------------------------------------------------------------------
void Grade2Solver::zeroPtrs ()
{
  _uu = 0;
  _z = 0;
  f_alpha = 0;
  f_mu = 0;
  f_rho = 0;
  f_w = 0;
  f_h = 0;
  _trans_funspace = 0;
  _stab_funspace = 0;
  _a_transport = 0;
  _L_transport = 0;
  _a_stab = 0;
  _L_stab = 0;
  _L_res = 0;
}
// ----------------------------------------------------------------------------
// end of file
