// Copyright (C) 2008-2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:  2008-11-18
// Last changed: 2009-12-08

#include "OldroydBSolver.h"

using namespace dolfin;

// ----------------------------------------------------------------------------
OldroydBSolver::OldroydBSolver ()
{
  zeroPtrs();
  // Set default parameters
  parameters.add("dt", 1.0e-2);
  parameters.add("T", 3.0);
  parameters.add("lam", 1.0);
  parameters.add("eta", 1.0);
  parameters.add("eta_e", 1.0);
}
// ----------------------------------------------------------------------------
OldroydBSolver::OldroydBSolver (const OldroydBSolver& that)
{
  throw std::runtime_error("Copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
OldroydBSolver::~OldroydBSolver ()
{
  deletePtrs();
}
// ----------------------------------------------------------------------------
OldroydBSolver& OldroydBSolver::operator= (const OldroydBSolver& that)
{
  throw std::runtime_error("Assignment is not implemented.");
  return *this;
}
// ----------------------------------------------------------------------------
void OldroydBSolver::F(GenericVector& b, const GenericVector& x)
{
  *_u = (*_uu)[0];
  *_p = (*_uu)[1];
  *_sigma = (*_uu)[2];

  logging(false);
  _assembler->assemble(b, *_L);

  for (unsigned int i = 0; i < _bcs.size(); i++)
    _bcs[i]->apply(b, x);
  logging(true);
}
// ----------------------------------------------------------------------------
void OldroydBSolver::J(GenericMatrix& A, const GenericVector& x)
{
  logging(false);
    _assembler->assemble(A, *_a);

  for (unsigned int i = 0; i < _bcs.size(); i++)
    _bcs[i]->apply(A);

  _assembler->assemble(*_b_proj, *_L_proj);

  _lu_solver.solve(*_A_proj, _D_h->vector(), *_b_proj);
  logging(true);
}
// ----------------------------------------------------------------------------
void OldroydBSolver::initialize_solver (Fluid& fluid, GenericFunction& f)
{

  // Get rid of old managed pointers
  deletePtrs();

  _fs = new OB_DEVSS::FunctionSpace(fluid.mesh());
  _fs_proj = new OB_StrainProj::FunctionSpace(fluid.mesh());
  SubSpace _fs_vel(*_fs, 0);
  SubSpace _fs_press(*_fs, 1);

  // Store references to unmanaged pointers
  fluid.setVelFunspace(_fs_vel);
  fluid.setPressFunspace(_fs_press);
  fluid.init();
  _u = &fluid.velocity();
  _p = &fluid.pressure();
  _mesh = &fluid.mesh();

  // Make boundary conditions,
  fluid.destroyBCs(_bcs);
  _bcs = fluid.createBCs(*_fs, false, true);

  // Create parameter functions
  f_dt = new Constant(parameters["dt"]);
  f_lam = new Lam(0.0);
  f_eta = new Constant(parameters["eta"]);
  f_eta_e = new Constant(parameters["eta_e"]);

  // Create discrete functions
  _uu_vec = new Vector(_fs->dofmap().global_dimension());
  _uu = new Function(*_fs, *_uu_vec);
  _D_h_vec = new Vector(_fs_proj->dofmap().global_dimension());
  _D_h = new Function(*_fs_proj, *_D_h_vec); //(*_fs_proj);
  _sigma_vec = new Vector(_fs_proj->dofmap().global_dimension());
  _sigma = new Function(*_fs_proj, *_sigma_vec);
  _sigma_n_vec = _sigma_vec;// new Vector(_fs_proj->dofmap().global_dimension());
  _sigma_n = new Function(*_fs_proj, *_sigma_n_vec);

  // Set forms
  _a = new OB_DEVSS::BilinearForm(*_fs, *_fs, *f_dt, *f_lam, *f_eta, *f_eta_e,
                                  *_sigma, *_u);
  _L = new OB_DEVSS::LinearForm(*_fs, *f_dt, *f_lam, *f_eta, *f_eta_e, f,
                                *_D_h, *_sigma, *_u, *_p, *_sigma_n);

  _a_proj = new OB_StrainProj::BilinearForm(*_fs_proj, *_fs_proj);
  _L_proj = new OB_StrainProj::LinearForm(*_fs_proj, *_u);

  // Create assembler and matrices
  _assembler = new Assembler;
  _A_proj = new Matrix;
  _b_proj = new Vector;

}
// ----------------------------------------------------------------------------
void OldroydBSolver::solve (Fluid& fluid, GenericFunction& f)
{
  // Initialize the solver
  initialize_solver(fluid, f);

  // Store variables for time and lambda stepping
  double dt = parameters["dt"];
  double T = parameters["T"];
  double lam_fin = parameters["lam"];
  double lam_curr = 0.0;
  double d_lam = 0.01;
  double lam_sgn = 1.0 ? lam_fin >= 0.0 : -1.0;

  // Assemble strain projection matrix
  logging(false);
  _assembler->assemble(*_A_proj, *_a_proj);
  logging(true);

  begin("Starting Oldroyd-B Solver.");
  while (lam_curr < lam_fin + DOLFIN_EPS) {
    info("--------------------------------------------------------------------------------");
    info("Lambda: %.2e <= %.2e", lam_curr, lam_fin);
    f_lam->setLam(lam_curr);

    begin("Starting Time stepper.");
    for (double t=0.0; t < T; t += dt) {
      info("------------------------------------------------------------------------------");
      info("T: %.3f", t);

      // Call the newton solver
      _newton_solver.solve(*this, _uu->vector());

      // Check to see if the time step has converged to a steady state
      if (_newton_solver.iteration() < 2) break;
    }
    end();

    // Double lambda or 0.1
    d_lam = lam_sgn * std::max(lam_curr, 0.1);
    if (lam_curr > lam_fin - DOLFIN_EPS)
      lam_curr = lam_fin + DOLFIN_EPS;
    else
      lam_curr = lam_sgn * std::min(fabs(lam_fin), fabs(lam_curr + d_lam));
  }
  end();

}
// ----------------------------------------------------------------------------
void OldroydBSolver::deletePtrs ()
{
  // Clean up the boundary conditions.
  for (uint i = 0; i < _bcs.size(); i++)
    delete _bcs[i];
  _bcs.resize(0);

  delete _fs;
  delete _fs_proj;
  delete _uu_vec;
  delete _uu;
  delete _sigma_vec;
  delete _sigma;
  delete _sigma_n_vec;
  delete _sigma_n;
  delete _D_h_vec;
  delete _D_h;
  delete f_dt;
  delete f_eta;
  delete f_eta_e;
  delete f_lam;
  delete _a;
  delete _L;
  delete _a_proj;
  delete _L_proj;
  delete _assembler;
  delete _A_proj;
  delete _b_proj;
  zeroPtrs();
}
// ----------------------------------------------------------------------------
void OldroydBSolver::zeroPtrs ()
{
  _fs = 0;
  _fs_proj = 0;
  _uu_vec = 0;
  _uu = 0;
  _sigma_vec = 0;
  _sigma = 0;
  _sigma_n_vec = 0;
  _sigma_n = 0;
  _D_h_vec = 0;
  _D_h = 0;
  f_dt = 0;
  f_eta = 0;
  f_eta_e = 0;
  f_lam = 0;
  _a = 0;
  _L = 0;
  _a_proj = 0;
  _L_proj = 0;
  _assembler = 0;
  _A_proj = 0;
  _b_proj = 0;
}
// ----------------------------------------------------------------------------
// end of file
