// Copyright (C) 2008-2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:  2008-11-18
// Last changed: 2009-12-08

#ifndef __OLDROYDBSOLVER_H
#define __OLDROYDBSOLVER_H

#include <dolfin.h>
#include "Fluid.h"

#include "forms/OldroydB/DEVSS/OB_DEVSS.h"
#include "forms/OldroydB/DEVSS/OB_StrainProj.h"

namespace dolfin
{

  class OldroydBSolver : public NonlinearProblem, public Variable
  {
  public:

    /// Create Oldroyd-B solver
    OldroydBSolver ();

    /// Copy constructor
    OldroydBSolver (const OldroydBSolver& that);

    /// Destructor
    ~OldroydBSolver ();

    /// Assignment Operator
    OldroydBSolver& operator= (const OldroydBSolver& that);

    void solve (Fluid& fluid, GenericFunction& f);

    /// Return Newton solver
    inline NewtonSolver& newton_solver () {return _newton_solver;}

    /// Function to compute F(u)
    void F (GenericVector& b, const GenericVector& x);

    /// Function to compute Jacobian matrix
    void J (GenericMatrix& A, const GenericVector& x);

  private:
    class Lam : public Expression
    {
    public:
      Lam(double lam) : Expression(), _lam(lam) {}
      void eval(double* vals, const std::vector<double>& input) const
      {
        vals[0] =_lam;
      }
      inline void setLam(double new_lam) {_lam = new_lam;}
    private:
      double _lam;
    };

    LUSolver _lu_solver;
    NewtonSolver _newton_solver;
    std::vector<const BoundaryCondition* > _bcs;

    // Managed pointers
    FunctionSpace* _fs;
    FunctionSpace* _fs_proj;
    Vector* _uu_vec;
    Function* _uu;
    Vector* _sigma_vec;
    Function* _sigma;
    Vector* _sigma_n_vec;
    Function* _sigma_n;
    Vector* _D_h_vec;
    Function* _D_h;
    GenericFunction* f_dt;
    GenericFunction* f_eta;
    GenericFunction* f_eta_e;
    Lam* f_lam;

    Form* _a;
    Form* _L;
    Form* _a_proj;
    Form* _L_proj;

    Assembler* _assembler;
    Matrix* _A_proj;
    Vector* _b_proj;

    // Unmanaged pointers
    Mesh* _mesh;
    Function* _u;
    Function* _p;

    // Private functions
    void deletePtrs ();
    void zeroPtrs ();

    void initialize_solver (Fluid& fluid, GenericFunction& f);

  };
}

#endif
