// Copyright (C) 2009 Patrick Riesen(riesen<at>vaw.baug.ethz.ch)
// Licensed under the GNU LGPL Version 3.0 or later.
//
// Modified by Andy R. Terrel, 2009
//
// First added:  2009-02-24
// Last changed: 2009-12-08


#include "PowerlawStokesSolver.h"

using namespace dolfin;

// ----------------------------------------------------------------------------
PowerlawStokesSolver::PowerlawStokesSolver () :
  NonlinearProblem(), Variable(),
  _V(0), _W(0), _S(0), _vel_V(0), _press_V(0),
  vv(0), eta(0), deta(0),
  uu(0), a(0), L(0), a_gamma(0), L_gamma(0)
{
  // Set default parameters
  parameters.add("kinematic_viscosity", 1.0);
  parameters.add("powerlaw_exponent", 1.0 );
  parameters.add("regularization_parameter", 1.e-5);
  parameters.add("continuation_steps", 4);

  parameters.add("convergence_tolerance", 1.0e-6);

  // Newton Solver parameters
  newtonsolver.parameters["convergence_criterion"] = "incremental";
  newtonsolver.parameters["maximum_iterations"] = 15;

}
// ----------------------------------------------------------------------------
PowerlawStokesSolver::PowerlawStokesSolver (const PowerlawStokesSolver& that)
{
  throw std::runtime_error("Copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
PowerlawStokesSolver::~PowerlawStokesSolver ()
{
  deletePtrs();
}
// ----------------------------------------------------------------------------
PowerlawStokesSolver& PowerlawStokesSolver::operator= (const PowerlawStokesSolver& that)
{
  throw std::runtime_error("Assignment is not implemented.");
  return *this;
}
//-----------------------------------------------------------------------------
void PowerlawStokesSolver::F (GenericVector& b, const GenericVector& x)
{
  logging(false);

  // store increment to fluid
  *_u = (*uu)[0];
  *_p = (*uu)[1];

  // call for viscosity update
  computeViscosity();

  // assemble F
  assemble(b, *L);

  for (uint i = 0; i < bcs.size(); i++)
    bcs[i]->apply(b, x);

  logging(true);
}
//-----------------------------------------------------------------------------
void PowerlawStokesSolver::J (GenericMatrix& A, const GenericVector& x)
{
  logging(false);

  // assemble J
  assemble(A, *a);

  for (uint i = 0; i < bcs.size(); i++)
    bcs[i]->apply(A);

  logging(true);
}
//-----------------------------------------------------------------------------
void PowerlawStokesSolver::createProblem (Fluid& fluid, GenericFunction& f)
{
  newton_max_iter = newtonsolver.parameters["maximum_iterations"];

  // set up function spaces
  _V = new PowerlawStokes::FunctionSpace(fluid.mesh());
  SubSpace v_fs(*_V, 0);
  SubSpace p_fs(*_V, 1);
  fluid.setVelFunspace(v_fs);
  fluid.setPressFunspace(p_fs);
  fluid.init();

  // addressing
  _u = &fluid.velocity();
  _p = &fluid.pressure();

  // Make boundary conditions
  bcs = fluid.createBCs(*_V, true, true);

  // function space for viscosity functions
  _W = new Gamma::FunctionSpace(fluid.mesh());

  /* ````````````````````````````````````````````````````````
     variable viscosity functions:
     eta = B*(0.5*tr(D(u),D(u)))**(1-n/2n)
     deta = eta' = B*(1-n)/2n*(0.5*tr(D(u),D(u)))**(1-n/2n - 1)
  */
  eta = new Eta(*_W); eta->vector() = 1.0;
  deta = new Eta(*_W); deta->vector().zero();

  /* ````````````````````````````````````````````````````````*/

  // solution func
  uu = new Function(*_V);
  uu->vector();

  // create the forms
  a = new PowerlawStokes::BilinearForm(*_V,*_V);
  PowerlawStokes::BilinearForm* _a_powstokes = dynamic_cast<PowerlawStokes::BilinearForm*>(a);
  _a_powstokes->uu = *uu;
  _a_powstokes->eta1 = *eta;
  _a_powstokes->eta2 = *deta;

  L = new PowerlawStokes::LinearForm(*_V);
  PowerlawStokes::LinearForm* _L_powstokes = dynamic_cast<PowerlawStokes::LinearForm*>(L);
  _L_powstokes->uu = *uu;
  _L_powstokes->eta1 = *eta;
  _L_powstokes->f = f;

  a_gamma = new Gamma::BilinearForm(*_W, *_W);
  L_gamma = new Gamma::LinearForm(*_W, *_u);

  vv = new Function(*_W); vv->vector().zero();

}
// ----------------------------------------------------------------------------
void PowerlawStokesSolver::solve (Fluid& fluid, GenericFunction& f)
{
  // create the forms and all
  createProblem(fluid,f);

  const double n = parameters["powerlaw_exponent"];
  const uint  T = parameters["continuation_steps"];

  // continuation when n>T0
  const double T0 = 1.8;

  // step size
  const double dt = (n-T0)/(T-1);

  // count
  uint newton_total_iter = 0;
  T_iter = 0; newton_iter = 0;

  if (n > T0) {
    nc = T0;
    // SOLVE WITH PARAMETER CONTINUATION
    for (T_iter = 0; T_iter < T; ++T_iter) {

      // require smaller tolerance during continuation
      if (T_iter < T-1)
        newtonsolver.parameters["relative_tolerance"] = 1.0e-3;
      else {
        double conv_tol = parameters["convergence_tolerance"];
        newtonsolver.parameters["relative_tolerance"] = conv_tol;
        newtonsolver.parameters["absolute_tolerance"] =  conv_tol;
      }

      logging(true);
      info("---------------------------------------------------------------------------------");
      info("N: %.3f (T = %d)", nc, T_iter);

      logging(false);

      // solve for exponent n = nc
      newtonsolver.solve(*this, uu->vector());

      logging(true);
      newton_total_iter = newton_total_iter + newton_iter;

      const bool report = newtonsolver.parameters["report"];
      if (report && fabs(nc-n) < 1e-5) {
        info("=================================================================================");
        info("Solve converged in total %d Newton iterations and %d Continuation steps",
                newton_total_iter,T);
        info("=================================================================================");
      }

      nc += dt;

    }
  }
  else {
    // SOLVE WITHOUT CONTINUATION
    newtonsolver.parameters["relative_tolerance"] = parameters["convergence_tolerance"];

    nc = n;
    newtonsolver.solve(*this, uu->vector());

    const bool report = newtonsolver.parameters["report"];
    if (report)
      info("Newton solve converged successfully.");
  }

  fluid.destroyBCs(bcs);
}
// ----------------------------------------------------------------------------
void PowerlawStokesSolver::computeViscosity ()
{
  const double B = parameters["kinematic_viscosity"];
  const double eps = parameters["regularization_parameter"];

  // compute the invariant trace(D(u)*D(u)) from
  // previous velocity solution
  VariationalProblem variable_viscosity(*a_gamma, *L_gamma);
  variable_viscosity.solve(*vv);

  // recreate viscosity dof vectors
  eta->set_vv(*vv,B,eps,nc);
  deta->set_Dvv(*vv,B,eps,nc);

  newton_iter = newtonsolver.iteration();
}
//-----------------------------------------------------------------------------
void PowerlawStokesSolver::computeStress (Fluid& fluid)
{
  // compute stresses from solution
  info("Requested stress computation:");

  _S = new PowerlawToStress::FunctionSpace(fluid.mesh());
  fluid.setStressFunspace(*_S);

  PowerlawToStress::BilinearForm a_stress(*_S,*_S);

  PowerlawToStress::LinearForm L_stress(*_S);
  L_stress.u = fluid.velocity();
  L_stress.p = fluid.pressure();
  L_stress.eta = *eta;

  VariationalProblem stressprojection(a_stress, L_stress);
  stressprojection.solve(fluid.stress());
}
//-----------------------------------------------------------------------------
void PowerlawStokesSolver::deletePtrs ()
{
  delete L_gamma;
  delete a_gamma;
  delete L;
  delete a;
  delete uu;
  delete deta;
  delete eta;
  delete vv;
  delete _press_V; _press_V = 0;
  delete _vel_V; _vel_V = 0;
  delete _S; _S = 0;
  delete _W; _W = 0;
  delete _V; _V = 0;
}
//-----------------------------------------------------------------------------
