// Copyright (C) 2009 Patrick Riesen(riesen<at>vaw.baug.ethz.ch)
// Licensed under the GNU LGPL Version 3.0 or later.
//
// Modified by Andy R. Terrel, 2009
//
// First added:  2009-02-24
// Last changed: 2009-12-08

#ifndef __POWERLAWSTOKESSOLVER_H
#define __POWERLAWSTOKESSOLVER_H

#include <dolfin.h>
#include "Fluid.h"

// See the FFC form files for information on the power-law model
#include "forms/powerlawStokes/TaylorHood/PowerlawStokes.h"
#include "forms/powerlawStokes/TaylorHood/Gamma.h"
#include "forms/powerlawStokes/TaylorHood/PowerlawToStress.h"

namespace dolfin
{

  // class for variable viscosity
  class Eta : public Function
  {
  public:
    Eta (FunctionSpace& W) : Function(W) {}
    //inline uint rank() const {return 0;}
    inline uint geometric_dimension () const {return 1;}
    void set_vv(const Function& vv, const double& B, const double& eps,
                const double nn)
    {
      // compute eta
      for (uint i = 0; i < vector().size(); i++)
        vector().setitem(i, B*pow(eps + vv.vector().getitem(i),
                                  (1.0-nn)/(2.0*nn)));
    }
    void set_Dvv(const Function& vv, const double& B, const double& eps,
                 const double nn)
    {
      // compute derivative of eta
      for (uint i = 0; i < vector().size(); i++)
        vector().setitem(i,
                         B*(1.0-nn)/(2.0*nn)*pow(eps + vv.vector().getitem(i),
                                                 (1.0-nn)/(2.0*nn)-1.0));
    }
  };

  // Solver class
  class PowerlawStokesSolver : public NonlinearProblem, public Variable
  {
  public:

    // Constructor
    PowerlawStokesSolver ();

    PowerlawStokesSolver (const PowerlawStokesSolver& that);

    // Destructor
    ~PowerlawStokesSolver ();

    PowerlawStokesSolver& operator= (const PowerlawStokesSolver& that);

    // compute L = F(u)
    void F(GenericVector& b, const GenericVector& x);

    // compute a = dF(u)du
    void J(GenericMatrix& A, const GenericVector& x);

    // solve the nonlinear problem
    void solve (Fluid& fluid, GenericFunction& f);

    void computeStress (Fluid& fluid);

  protected:

    // protected functions
    void createProblem (Fluid& fluid, GenericFunction& f);
    void computeViscosity ();
    void deletePtrs ();

  private:

    // newton solve utilities
    uint newton_max_iter;
    uint T_iter;
    uint newton_iter;
    double nc;
    NewtonSolver newtonsolver;
    std::vector<const BoundaryCondition* > bcs;

    // function spaces
    FunctionSpace* _V;
    FunctionSpace* _W;
    FunctionSpace*_S;

    SubSpace* _vel_V;
    SubSpace* _press_V;

    // variable viscosity funcs
    Function* vv;
    Eta* eta;
    Eta* deta;

    // solution func
    Function* uu;

    // the forms
    Form* a;
    Form* L;
    Form* a_gamma;
    Form* L_gamma;

    // adressers to fluid.vel/press
    Function* _u;
    Function* _p;

  };

}
#endif
