// Copyright (C) 2008--2009 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:   2008-09-12
// Last modified: 2009-12-08

#include "StokesSolver.h"

using namespace dolfin;

// ----------------------------------------------------------------------------
StokesSolver::StokesSolver () :
  _funspace(0), _a_stokes(0), _L_stokes(0), _a_div_proj(0), _L_div_proj(0)
{
  // Set default parameters
  parameters.add("velocity_disc","Taylor-Hood");
  parameters.add("SV_penalty_r",1.0e5);
  parameters.add("SV_penalty_rho",1.0e5);
  parameters.add("SV_max_iter",100);
}
// ----------------------------------------------------------------------------
StokesSolver::StokesSolver (const StokesSolver& that)
{
  throw std::runtime_error("Copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
StokesSolver::~StokesSolver ()
{
  deletePtrs();
}
// ----------------------------------------------------------------------------
StokesSolver& StokesSolver::operator= (const StokesSolver& that)
{
  throw std::runtime_error("Assignment is not implemented.");
  return *this;
}
// ----------------------------------------------------------------------------
void StokesSolver::solve (Fluid& fluid, GenericFunction& f)
{
  // Get rid of old data, and get the new discretization method
  deletePtrs();
  const std::string disc = parameters["velocity_disc"];

  // Make boundary conditions, if it is a mixed method put
  // the bc's on the velocity subsystem.
  std::vector<const BoundaryCondition* > bcs;

  info("Starting Stokes solver.");
  // Set forms and call the particular solution for each method.
  if (disc == "Taylor-Hood") {
    _funspace = new P2THStokes::FunctionSpace(fluid.mesh());
    SubSpace v_fs(*_funspace, 0);
    SubSpace p_fs(*_funspace, 1);
    fluid.setVelFunspace(v_fs);
    fluid.setPressFunspace(p_fs);
    fluid.init();
    bcs = fluid.createBCs(*_funspace, true, true);
    _a_stokes = new P2THStokes::BilinearForm(*_funspace, *_funspace);
    _L_stokes = new P2THStokes::LinearForm(*_funspace, f);
    mixedMethodSolve(fluid, bcs, f);
  } else if (disc == "Scott-Vogelius") {
    _funspace = new P4SVStokes::FunctionSpace(fluid.mesh());
    P4VecDiv::FunctionSpace press_fs(fluid.mesh());
    fluid.setVelFunspace(*_funspace);
    fluid.setPressFunspace(press_fs);
    fluid.init();
    bcs = fluid.createBCs(*_funspace, false);
    Constant f_rho(parameters["SV_penalty_rho"]);
    Function f_w(*_funspace);
    _a_stokes = new P4SVStokes::BilinearForm(*_funspace, *_funspace, f_rho);
    _L_stokes = new P4SVStokes::LinearForm(*_funspace, f, f_w);
    _a_div_proj = new P4VecDiv::BilinearForm(*fluid.press_funspace(),
                                             *fluid.press_funspace());
    _L_div_proj = new P4VecDiv::LinearForm(*fluid.press_funspace(), f_w);
    scottVogeliusSolve(fluid, bcs, f, f_w);
  } else {
    throw std::runtime_error("Unimplemented Stokes discretization");
  }

  // Clean up the boundary conditions.
  fluid.destroyBCs(bcs);
}
// ----------------------------------------------------------------------------
void StokesSolver::mixedMethodSolve (Fluid& fluid,
                                     std::vector<const BoundaryCondition* >& bcs,
				     GenericFunction& f)
{
  // Solve the PDE
  VariationalProblem pde_stokes(*_a_stokes, *_L_stokes, bcs);
  pde_stokes.solve(fluid.velocity(),fluid.pressure());
}
// ----------------------------------------------------------------------------
void StokesSolver::scottVogeliusSolve (Fluid& fluid,
                                       std::vector<const BoundaryCondition* >& bcs,
				       GenericFunction& f, Function& f_w)
{
  const double r = parameters["SV_penalty_r"];
  const int max_iter = parameters["SV_max_iter"];

  // Initialize temporary function w
  f_w.vector().zero();

  // Create pde and update until ||div(u)|| < eps
  VariationalProblem pde_stokes(*_a_stokes, *_L_stokes, bcs);
  P4VecDivL2::Functional div_func(fluid.mesh(), fluid.velocity());
  double div_u = 1.0;
  for (int i = 0; i < max_iter && div_u > DOLFIN_SQRT_EPS; i++) {
    logging(false);
    pde_stokes.solve(fluid.velocity());
    logging(true);
    f_w.vector().axpy(r,fluid.velocity().vector());
    div_u = sqrt(fabs(assemble(div_func)));
    info("Scott Vogelius iteration %d, ||div(u)|| = %.2e", i, div_u);
  }

  // Project div(w) to get the pressure
  VariationalProblem div_proj(*_a_div_proj, *_L_div_proj);
  logging(false);
  div_proj.solve(fluid.pressure());
  logging(true);

}
// ----------------------------------------------------------------------------
void StokesSolver::deletePtrs ()
{
  delete _funspace; _funspace = 0;
  delete _a_stokes; _a_stokes = 0;
  delete _L_stokes; _L_stokes = 0;
  delete _a_div_proj; _a_div_proj = 0;
  delete _L_div_proj; _L_div_proj = 0;
}
// ----------------------------------------------------------------------------
// end of file
